// dictionary.c
// 
// A simple and robust dynamic dictionary implementation in ANSI C using pages and hash values
// for faster access to the data.
// 
// This code is very easily portable to C++.


#include "dictionary.h"


// internal functions
unsigned long Dictionary_ComputeHashValueForKey (char *key);


dictionary_t *Dictionary_CreateDictionary (int size)
{
   // this function allocates a new dictionary object of given size and returns it. If you do
   // not know in advance (roughly) the number of entries in the dictionary, give size = 0.

   dictionary_t *dictionary;
   int index;

   // if no size was specified, allocate space for one page
   if (size < DICTIONARY_PAGE_SIZE)
      size = DICTIONARY_PAGE_SIZE;

   dictionary = (dictionary_t *) malloc (sizeof (dictionary_t));

   dictionary->entry_count = 0; // no entries in dictionary yet
   dictionary->size = size; // dictionary can currently hold size entries

   // allocate space for the dictionary entries and zero all the crap out
   dictionary->entries = (dictionary_entry_t *) malloc (size * sizeof (dictionary_entry_t));
   memset (dictionary->entries, 0, size * sizeof (dictionary_entry_t));

   // for each entry in the dictionary...
   for (index = 0; index < size; index++)
   {
      // mallocate the minimal space for key and value and zero them out
      dictionary->entries[index].key = (char *) malloc (1 * sizeof (char));
      dictionary->entries[index].key[0] = 0;
      dictionary->entries[index].value = (char *) malloc (1 * sizeof (char));
      dictionary->entries[index].value[0] = 0;
   }

   return (dictionary); // finished, return a pointer to the dictionary object
}


void Dictionary_DestroyDictionary (dictionary_t *dictionary)
{
   // frees a dictionary object and all memory associated to it.

   int index;

   if (dictionary == NULL)
      return; // reliability check

   // for each entry in the dictionary...
   for (index = 0; index < dictionary->size; index++)
   {
      if (dictionary->entries[index].key != NULL)
         free (dictionary->entries[index].key); // free the key
      dictionary->entries[index].key = NULL;

      if (dictionary->entries[index].value != NULL)
         free (dictionary->entries[index].value); // free the value
      dictionary->entries[index].value = NULL;
   }

   if (dictionary->entries != NULL)
      free (dictionary->entries); // free the entries array
   dictionary->entries = NULL;

   if (dictionary != NULL)
      free (dictionary); // and finally, free the dictionary itself
   dictionary = NULL;

   return; // finished
}


char *Dictionary_ReadKey (dictionary_t *dictionary, char *key, char *def)
{
   // this function locates a key in a dictionary and returns a pointer to its value, or the
   // passed 'def' pointer if no such key can be found in dictionary. The returned char pointer
   // points to data internal to the dictionary object, do not try to free or modify it.

   unsigned long hash;
   int index;

   hash = Dictionary_ComputeHashValueForKey (key); // get the hash value for key

   // cycle through all entries in the dictionary
   for (index = 0; index < dictionary->size; index++)
   {
      if (dictionary->entries[index].key[0] == 0)
         continue; // skip empty slots

      // compare hash AND string, to avoid hash collisions
      if ((hash == dictionary->entries[index].hash) && (strcmp (key, dictionary->entries[index].key) == 0))
         return (dictionary->entries[index].value); // return the value if we find the key
   }

   return (def); // else return the default value
}


void Dictionary_WriteKey (dictionary_t *dictionary, char *key, char *value)
{
   // sets a value in a dictionary. If the given key is found in the dictionary, the associated
   // value is replaced by the provided one. If the key cannot be found in the dictionary, it
   // is added to it.

   unsigned long hash;
   int index;

   if ((dictionary == NULL) || (key == NULL))
      return; // reliability check
   
   hash = Dictionary_ComputeHashValueForKey (key); // compute hash for this key

   // is the dictionary NOT empty ?
   if (dictionary->entry_count > 0)
   {
      // for each entry in the dictionary...
      for (index = 0; index < dictionary->size; index++)
      {
         if (dictionary->entries[index].key[0] == 0)
             continue; // skip empty slots

         // does that key have the same hash value AND the same name ?
         if ((hash == dictionary->entries[index].hash) && (strcmp (key, dictionary->entries[index].key) == 0))
         {
            // we've found the right key: modify its value and return

            // have we provided a valid value ?
            if (value != NULL)
            {
               // yes, so allocate/reallocate and copy the value string
               dictionary->entries[index].value = (char *) realloc (dictionary->entries[index].value, strlen (value) * sizeof (char));
               strcpy (dictionary->entries[index].value, value); // copy the value string
            }
            else
            {
               // no, so reallocate and reset the value string
               dictionary->entries[index].value = (char *) realloc (dictionary->entries[index].value, 1 * sizeof (char));
               dictionary->entries[index].value[0] = 0; // reset the value string
            }

            return; // value has been modified: return
         }
      }
   }

   // here either the dictionary was empty, or we couldn't find a similar value: add a new one

   // see if dictionary needs to grow
   if (dictionary->entry_count == dictionary->size)
   {
      dictionary->size += DICTIONARY_PAGE_SIZE; // increase dictionary size

      // mallocate some more space for the dictionary and zero all that crap out
      dictionary->entries = (dictionary_entry_t *) realloc (dictionary->entries, dictionary->size * sizeof (dictionary_entry_t)); // alloc 1 more page
      memset (&dictionary->entries[dictionary->entry_count], 0, DICTIONARY_PAGE_SIZE * sizeof (dictionary_entry_t)); // zero out

      // for each additional entry in the dictionary...
      for (index = dictionary->entry_count; index < dictionary->size; index++)
      {
         // mallocate the minimal space for key and value and zero them out
         dictionary->entries[index].key = (char *) malloc (1 * sizeof (char));
         dictionary->entries[index].key[0] = 0;
         dictionary->entries[index].value = (char *) malloc (1 * sizeof (char));
         dictionary->entries[index].value[0] = 0;
      }
   }

   // cycle through all entries in the dictionary...
   for (index = 0; index < dictionary->size; index++)
      if (dictionary->entries[index].key[0] == 0)
         break; // and stop at the first empty one we find

   // reallocate and copy the new key

   dictionary->entries[index].key = (char *) realloc (dictionary->entries[index].key, strlen (key) * sizeof (char));
   strcpy (dictionary->entries[index].key, key); // copy the key string...
   dictionary->entries[index].hash = hash; // ...and store the hash value

   // have we provided a valid value ?
   if (value != NULL)
   {
      // yes, so reallocate and copy the value string
      dictionary->entries[index].value = (char *) realloc (dictionary->entries[index].value, strlen (value) * sizeof (char));
      strcpy (dictionary->entries[index].value, value); // copy the value string
   }
   else
   {
      // no, so reallocate and reset the value string
      dictionary->entries[index].value = (char *) realloc (dictionary->entries[index].value, 1 * sizeof (char));
      dictionary->entries[index].value[0] = 0; // reset the value string
   }

   dictionary->entry_count++; // there is one more entry in the dictionary now
   return;
}


void Dictionary_DeleteKey (dictionary_t *dictionary, char *key)
{
   // this function deletes a key in a dictionary. Nothing is done if the key cannot be found.

   unsigned long hash;
   int index;

   hash = Dictionary_ComputeHashValueForKey (key); // get the hash value for key

   // cycle through all entries in the dictionary...
   for (index = 0; index < dictionary->size; index++)
   {
      if (dictionary->entries[index].key[0] == 0)
         continue; // skip empty slots

      // compare hash AND string, to avoid hash collisions
      if ((hash == dictionary->entries[index].hash) && (strcmp (key, dictionary->entries[index].key) == 0))
         break; // break as soon as we've found the key we want
   }

   if (index == dictionary->size)
      return; // if key was not found, just return

   // clear the key, the hash and its value data
   dictionary->entries[index].key = (char *) realloc (dictionary->entries[index].key, 1 * sizeof (char));
   dictionary->entries[index].key[0] = 0;
   dictionary->entries[index].hash = 0;
   dictionary->entries[index].value = (char *) realloc (dictionary->entries[index].value, 1 * sizeof (char));
   dictionary->entries[index].value[0] = 0;

   dictionary->entry_count--; // there is one entry less in the dictionary now
   return;
}


unsigned long Dictionary_ComputeHashValueForKey (char *key)
{
   // Compute the hash key for a string. This hash function has been taken from an article in
   // Dr Dobbs Journal. This is normally a collision-free function, distributing keys evenly.
   // The key is stored anyway in the struct so that collision can be avoided by comparing the
   // key itself in last resort. THIS FUNCTION IS TO BE USED INTERNALLY ONLY!

   unsigned long hash;
   int length;
   int index;

   hash = 0;
   length = strlen (key);

   // for each character of the string...
   for (index = 0; index < length; index++)
   {
      hash += (unsigned long) key[index]; // take it in account and compute the hash value
      hash += (hash << 10);
      hash ^= (hash >> 6);
   }

   hash += (hash << 3); // finalize hashing (what the hell does it do, I have no clue)
   hash ^= (hash >> 11);
   hash += (hash << 15);

   return (hash); // and return the hash value
}
