// inifile.c
// 
// A simple and robust INI reader/writer implementation in ANSI C using a fast underlying
// dictionary structure.
// 
// This code is very easily portable to C++.


#include "inifile.h"


#define INI_SECTION_SEPARATOR '#'


char *INIFile_default_section = "__default";


int INIFile_GetNumberOfSections (dictionary_t *dictionary)
{
   // this function returns the number of sections found in a dictionary. The test to recognize
   // sections is done on the string stored in the dictionary: a section name is given as
   // "section" whereas a key is stored as "section#key", thus the test looks for entries that
   // do NOT contain a '#'. This clearly fails in the case a section name contains a '#',
   // but this should simply be avoided. This function returns -1 in case of error.

   int section_count;
   int index;

   if (dictionary == NULL)
      return (-1); // reliability check

   section_count = 0; // no sections found yet

   // for each entry in the dictionary...
   for (index = 0; index < dictionary->size; index++)
   {
      if (dictionary->entries[index].key[0] == 0)
         continue; // skip empty slots

      if (strchr (dictionary->entries[index].key, INI_SECTION_SEPARATOR) == NULL)
         section_count++; // if this entry has NO section separator, then it's a section name
   }

   return (section_count); // return the section count we found
}


char *INIFile_GetSectionName (dictionary_t *dictionary, int section_index)
{
   // this function locates the n-th section in a dictionary and returns its name as a pointer
   // to a string statically allocated inside the dictionary. Do not free or modify the
   // returned string! This function returns NULL in case of error.

   int sections_found;
   int index;

   if ((dictionary == NULL) || (section_index < 0))
      return (NULL); // reliability check

   sections_found = 0; // no sections found yet

   // for each entry in the dictionary...
   for (index = 0; index < dictionary->size; index++)
   {
      if (dictionary->entries[index].key[0] == 0)
         continue; // skip empty slots

      // is this entry a section name ?
      if (strchr (dictionary->entries[index].key, INI_SECTION_SEPARATOR) == NULL)
      {
         // is it the section we want ?
         if (sections_found == section_index)
            return (dictionary->entries[index].key); // then return its name

         sections_found++; // we found one section more
      }
   }

   return (INIFile_default_section); // section was not found, return the default section name
}


bool INIFile_ReadEntryAsBool (dictionary_t *dictionary, char *section, char *entry, bool default_value)
{
   // this function queries a dictionary for a key. A key as read from an ini file is given as
   // "section#key". If the key cannot be found, the notfound value is returned. A true boolean
   // is found if a string starting with either 'y', 'Y', 't', 'T' or '1' is matched. A false
   // boolean is found if a string starting with 'n', 'N', 'f', 'F' or '0' is matched.

   char *value;

   // get the value as a string first
   value = INIFile_ReadEntryAsString (dictionary, section, entry, DICTIONARY_INVALID_KEY);

   if (value == DICTIONARY_INVALID_KEY)
      return (default_value); // if the entry could not be read, return the default value

   // decide of the boolean's value by looking at the first character
   if ((toupper (value[0]) == 'T') || (toupper (value[0]) == 'Y') || (value[0] == '1'))
     return (true); // boolean value is TRUE
   else if ((toupper (value[0]) == 'F') || (toupper (value[0]) == 'N') || (value[0] == '0'))
     return (false); // boolean value is FALSE

   return (default_value); // boolean value is undefined, return default value
}


long INIFile_ReadEntryAsLong (dictionary_t *dictionary, char *section, char *entry, long default_value)
{
   // this function queries a dictionary for a key. A key as read from an ini file is given as
   // "section#key". If the key cannot be found, the notfound value is returned.

   char *value;

   // get the value as a string first
   value = INIFile_ReadEntryAsString (dictionary, section, entry, DICTIONARY_INVALID_KEY);

   if (value == DICTIONARY_INVALID_KEY)
      return (default_value); // if the entry could not be read, return the default value

   return (atoi (value)); // else convert the value to a long integer and return it
}


double INIFile_ReadEntryAsDouble (dictionary_t *dictionary, char *section, char *entry, double default_value)
{
   // this function queries a dictionary for a key. A key as read from an ini file is given as
   // "section#key". If the key cannot be found, the notfound value is returned.

   char *value;

   // get the value as a string first
   value = INIFile_ReadEntryAsString (dictionary, section, entry, DICTIONARY_INVALID_KEY);

   if (value == DICTIONARY_INVALID_KEY)
      return (default_value); // if the entry could not be read, return the default value

   return (atof (value)); // else convert the value to a double precision number and return it
}


char *INIFile_ReadEntryAsString (dictionary_t *dictionary, char *section, char *entry, char *default_value)
{
   // this function queries a dictionary for a key. A key as read from an ini file is given as
   // "section#key". If the key cannot be found, the pointer passed as 'def' is returned. The
   // returned char pointer is pointing to a string allocated in the dictionary, do not free
   // or modify it.

   char key[256];
   int length;
   int i;
   char *value;

   if ((dictionary == NULL) || (key == NULL))
      return (default_value); // reliability check

   if (section == NULL)
      section = INIFile_default_section; // if no section was provided, use empty section name

   // if we were given an entry, build a key as section#entry
   if (entry != NULL)
   {
      sprintf (key, "%s%c%s", section, INI_SECTION_SEPARATOR, entry); // compose the key
      length = strlen (key); // get the key string length

      // for each character in the string after the section separator...
      for (i = strlen (section) + 1; i < length; i++)
         key[i] = tolower (key[i]); // convert it to lowercase
      key[i] = 0; // terminate the string
   }

   // else it must be a section name
   else
      strcpy (key, section); // copy the name into the key

   value = Dictionary_ReadKey (dictionary, key, default_value); // query the dictionary...
   return (value); // ...and return the value
}


void INIFile_WriteEntryAsBool (dictionary_t *dictionary, char *section, char *entry, bool value)
{
   // sets an entry in a dictionary. If the given entry can be found in the dictionary, it is
   // modified to contain the provided value. If it cannot be found, it is created.

   char value_as_string[256];

   // build the boolean's value equivalent string
   strcpy (value_as_string, (value ? "true" : "false")); // value is either TRUE or FALSE

   INIFile_WriteEntryAsString (dictionary, section, entry, value_as_string); // write the new entry
   return; // finished
}


void INIFile_WriteEntryAsLong (dictionary_t *dictionary, char *section, char *entry, long value)
{
   // sets an entry in a dictionary. If the given entry can be found in the dictionary, it is
   // modified to contain the provided value. If it cannot be found, it is created.

   char value_as_string[256];

   // build the long integer value equivalent string (use printf facility)
   sprintf (value_as_string, "%d", value);

   INIFile_WriteEntryAsString (dictionary, section, entry, value_as_string); // write the new entry
   return; // finished
}


void INIFile_WriteEntryAsDouble (dictionary_t *dictionary, char *section, char *entry, double value)
{
   // sets an entry in a dictionary. If the given entry can be found in the dictionary, it is
   // modified to contain the provided value. If it cannot be found, it is created.

   char value_as_string[256];

   // build the long integer value equivalent string (use printf facility)
   sprintf (value_as_string, "%g", value);

   INIFile_WriteEntryAsString (dictionary, section, entry, value_as_string); // write the new entry
   return; // finished
}


void INIFile_WriteEntryAsString (dictionary_t *dictionary, char *section, char *entry, char *value)
{
   // sets an entry in a dictionary. If the given entry can be found in the dictionary, it is
   // modified to contain the provided value. If it cannot be found, it is created.

   char key[256];
   int length;
   int i;

   if (dictionary == NULL)
      return; // reliability check

   if (section == NULL)
      section = INIFile_default_section; // if no section was provided, use empty section name

   // if we were given an entry, build a key as section#entry
   if (entry != NULL)
   {
      Dictionary_WriteKey (dictionary, section, NULL); // create the section if it doesn't exist

      sprintf (key, "%s%c%s", section, INI_SECTION_SEPARATOR, entry); // compose the key
      length = strlen (key); // get the key string length

      // for each character in the string after the section separator...
      for (i = strlen (section) + 1; i < length; i++)
         key[i] = tolower (key[i]); // convert it to lowercase
      key[i] = 0; // terminate the string
   }

   // else it must be a section name
   else
      strcpy (key, section); // copy the name into the key

   Dictionary_WriteKey (dictionary, key, value); // write the new key in the dictionary
   return; // finished
}


void INIFile_DeleteEntry (dictionary_t *dictionary, char *section, char *entry)
{
   // deletes an entry in the dictionary

   char key[256];
   int length;
   int i;

   if ((dictionary == NULL) || (entry == NULL))
      return; // reliability check

   if (section == NULL)
      section = INIFile_default_section; // if no section was provided, use empty section name

   // if we were given an entry, build a key as section#entry
   if (entry != NULL)
   {
      sprintf (key, "%s%c%s", section, INI_SECTION_SEPARATOR, entry); // compose the key
      length = strlen (key); // get the key string length

      // for each character in the string after the section separator...
      for (i = strlen (section) + 1; i < length; i++)
         key[i] = tolower (key[i]); // convert it to lowercase
      key[i] = 0; // terminate the string
   }

   // else it must be a section name
   else
      strcpy (key, section); // copy the name into the key

   Dictionary_DeleteKey (dictionary, key);
   return;
}


void INIFile_DeleteSection (dictionary_t *dictionary, char *section)
{
   // deletes a whole INI section in the dictionary

   char key[256];
   int length;
   int i;

   if (dictionary == NULL)
      return; // reliability check

   if (section == NULL)
      section = INIFile_default_section; // if no section was provided, use empty section name

   sprintf (key, "%s%c", section, INI_SECTION_SEPARATOR); // compose the key
   length = strlen (key); // get the key string length

   // for each entry in the dictionary...
   for (i = 0; i < dictionary->size; i++)
   {
      if (dictionary->entries[i].key[0] == 0)
         continue; // skip empty slots

      // does this entry belong to the section we want ?
      if (strncmp (dictionary->entries[i].key, key, length) == 0)
         Dictionary_DeleteKey (dictionary, dictionary->entries[i].key); // yes, delete it
   }

   Dictionary_DeleteKey (dictionary, section); // and finally delete the section name itself
   return;
}


void INIFile_LoadINIFile (char *filename, dictionary_t *dictionary)
{
   // this is the parser for ini files. This function is called, providing the name of the file
   // to be read. It returns a dictionary object that should not be accessed directly, but
   // through accessor functions instead.

   FILE *fp;
   char line_buffer[1024];
   char section[256];
   char entry[256];
   char value[256];
   int fieldstart;
   int fieldstop;
   int length;
   int i;

   if ((dictionary == NULL) || (filename == NULL))
      return; // reliability check

   // try to open the INI file in ASCII read-only mode
   fp = fopen (filename, "r");
   if (fp == NULL)
      return; // cancel if file not found

   // set the default section for orphaned entries and add it to the dictionary
   strcpy (section, INIFile_default_section);
   INIFile_WriteEntryAsString (dictionary, INIFile_default_section, NULL, NULL);

   // read line per line...
   while (fgets (line_buffer, 1024, fp) != NULL)
   {
      length = strlen (line_buffer); // get line length

      while ((length > 0) && ((line_buffer[length - 1] == '\n') || (line_buffer[length - 1] == '\r')))
         length--; // discard trailing newlines

      fieldstart = 0; // let's now strip leading blanks
      while ((fieldstart < length) && isspace (line_buffer[fieldstart]))
         fieldstart++; // ignore any tabs or spaces, going forward from the start

      fieldstop = length - 1; // let's now strip trailing blanks
      while ((fieldstop >= 0) && isspace (line_buffer[fieldstop]))
         fieldstop--; // ignore any tabs or spaces, going backwards from the end

      for (i = fieldstart; i <= fieldstop; i++)
         line_buffer[i - fieldstart] = line_buffer[i]; // recopy line buffer without the spaces
      line_buffer[i - fieldstart] = 0; // and terminate the string

      if ((line_buffer[0] == ';') || (line_buffer[0] == '#') || (line_buffer[0] == 0))
         continue; // skip comment lines

      // is it a valid section name ?
      if (sscanf (line_buffer, "[%[^]]", section) == 1)
      {
         length = strlen (section); // get the section string length

         fieldstart = 0; // let's now strip leading blanks
         while ((fieldstart < length) && isspace (section[fieldstart]))
            fieldstart++; // ignore any tabs or spaces, going forward from the start

         fieldstop = length - 1; // let's now strip trailing blanks
         while ((fieldstop >= 0) && isspace (section[fieldstop]))
            fieldstop--; // ignore any tabs or spaces, going backwards from the end

         for (i = fieldstart; i <= fieldstop; i++)
            section[i - fieldstart] = tolower (section[i]); // recopy section name w/out spaces
         section[i - fieldstart] = 0; // and terminate the string

         INIFile_WriteEntryAsString (dictionary, section, NULL, NULL); // add to dictionary
      }

      // else is it a valid entry/value pair ?
      else if ((sscanf (line_buffer, "%[^=] = \"%[^\"]\"", entry, value) == 2)
               || (sscanf (line_buffer, "%[^=] = '%[^\']'", entry, value) == 2)
               || (sscanf (line_buffer, "%[^=] = %[^;#]", entry, value) == 2))
      {
         length = strlen (entry); // get the entry string length

         fieldstart = 0; // let's now strip leading blanks
         while ((fieldstart < length) && isspace (entry[fieldstart]))
            fieldstart++; // ignore any tabs or spaces, going forward from the start

         fieldstop = length - 1; // let's now strip trailing blanks
         while ((fieldstop >= 0) && isspace (entry[fieldstop]))
            fieldstop--; // ignore any tabs or spaces, going backwards from the end

         for (i = fieldstart; i <= fieldstop; i++)
            entry[i - fieldstart] = tolower (entry[i]); // recopy entry name w/out spaces
         entry[i - fieldstart] = 0; // and terminate the string

         length = strlen (value); // get the value string length

         fieldstart = 0; // let's now strip leading blanks
         while ((fieldstart < length) && isspace (value[fieldstart]))
            fieldstart++; // ignore any tabs or spaces, going forward from the start

         fieldstop = length - 1; // let's now strip trailing blanks
         while ((fieldstop >= 0) && isspace (value[fieldstop]))
            fieldstop--; // ignore any tabs or spaces, going backwards from the end

         for (i = fieldstart; i <= fieldstop; i++)
            value[i - fieldstart] = value[i]; // recopy entry name w/out spaces
         value[i - fieldstart] = 0; // and terminate the string

         // sscanf cannot handle "" or '' as empty value, this is done here
         if ((strcmp (value, "\"\"") == 0) || (strcmp (value, "''") == 0))
            value[0] = 0; // empty string

         INIFile_WriteEntryAsString (dictionary, section, entry, value); // add to dictionary
      }
   }

   fclose (fp); // finished, close the file

   return; // and return
}


void INIFile_SaveINIFile (char *filename, dictionary_t *dictionary)
{
   // this function dumps a given dictionary into a loadable ini file.

   FILE *fp;
   char key[256];
   int section_index;
   char *section_name;
   int section_count;
   int index;
   int length;

   if ((dictionary == NULL) || (filename == NULL))
      return; // reliability check

   // try to open the INI file in ASCII write mode
   fp = fopen (filename, "w");
   if (fp == NULL)
      return; // cancel if unable to open file

   // print the INI file name as a comment
   fprintf (fp, "# %s\n", filename);

   // get the number of sections there are in this INI dictionary
   section_count = INIFile_GetNumberOfSections (dictionary);

   // for each section...
   for (section_index = 0; section_index < section_count; section_index++)
   {
      section_name = INIFile_GetSectionName (dictionary, section_index); // read section name

      // is it the default section ?
      if (strcmp (section_name, INIFile_default_section) == 0)
         fprintf (fp, "\n"); // don't put the default section's name in the INI file
      else
         fprintf (fp, "\n[%s]\n", section_name); // dump all other sections into the INI file

      // build the section identifier to be used when looking up the dictionary
      sprintf (key, "%s%c", section_name, INI_SECTION_SEPARATOR);
      length = strlen (key);

      // then for each entry in the dictionary...
      for (index = 0; index < dictionary->size; index++)
      {
         if (dictionary->entries[index].key[0] == 0)
            continue; // skip empty slots

         // does this key belong to the section we want ? if so, dump it
         if (strncmp (dictionary->entries[index].key, key, length) == 0)
            fprintf (fp, "%s = %s\n", dictionary->entries[index].key + length, dictionary->entries[index].value);
      }
   }

   fclose (fp); // finished, close the file

   return; // and return
}
